from typing import Callable, List, Mapping, Optional, Set

import drepr.models.sm as drepr_sm
import sm.outputs.semantic_model as O
from sm.misc.funcs import assert_not_null
from sm.namespaces.prelude import KnowledgeGraphNamespace

from sand.config import AppConfig
from sand.helpers.namespace import NamespaceService
from sand.models.ontology import OntProperty, OntPropertyDataType

# mapping from predefined datatypes to D-REPR datatype
datatype_mapping: Mapping[OntPropertyDataType, Optional[drepr_sm.DataType]] = {
    "globe-coordinate": drepr_sm.DataType(
        "http://www.opengis.net/ont/geosparql#wktLiteral",
        {"geo": "http://www.opengis.net/ont/geosparql#"},
    ),
    "url": drepr_sm.PredefinedDataType.xsd_anyURI.value,
    "entity": drepr_sm.PredefinedDataType.drepr_uri.value,
    "string": drepr_sm.PredefinedDataType.xsd_string.value,
    "integer-number": drepr_sm.PredefinedDataType.xsd_int.value,
    "decimal-number": drepr_sm.PredefinedDataType.xsd_decimal.value,
    "datetime": drepr_sm.PredefinedDataType.xsd_dateTime.value,
}


def get_drepr_sm(
    sm: O.SemanticModel,
    kgns: NamespaceService,
    kgns_prefixes: dict[str, str],
    ontprop_ar: Mapping[str, OntProperty],
    ident_props: set[str],
    get_attr_id: Callable[[int], str],
    get_ent_attr_id: Callable[[int], str],
) -> drepr_sm.SemanticModel:
    """Convert sm model into drepr model.

    Args:
        sm: the semantic model we want to convert
        kgns: the knowledge graph namespace
        kgns_prefixes: the prefixes of the knowledge graph namespace
        ontprop_ar: mapping from the id to ontology property
        ident_props: list of properties that telling a data node contains entities (e.g., rdfs:label)
        get_attr_id: get attribute id from column index
        get_ent_attr_id: for each entity column, to generate url, we create an extra attribute containing the entity uri, this function get its id based on the column index
    """
    nodes = {}
    edges = {}

    for node in sm.nodes():
        if isinstance(node, O.ClassNode):
            nodes[str(node.id)] = drepr_sm.ClassNode(
                node_id=str(node.id), label=node.rel_uri
            )
        elif isinstance(node, O.DataNode):
            # find data type of this node, when they have multiple data types
            # that do not agree with each other, we don't set the datatype
            # usually, that will be displayed from the UI so users know that

            datatypes: Set[OntPropertyDataType] = {
                assert_not_null(ontprop_ar[kgns.uri_to_id(inedge.abs_uri)]).datatype
                for inedge in sm.in_edges(node.id)
            }
            datatype = (
                datatype_mapping.get(list(datatypes)[0], None)
                if len(datatypes) == 1
                else None
            )

            nodes[str(node.id)] = drepr_sm.DataNode(
                node_id=str(node.id),
                attr_id=get_attr_id(node.col_index),
                data_type=datatype,
            )
        elif isinstance(node, O.LiteralNode):
            if node.datatype == O.LiteralNodeDataType.Entity:
                datatype = drepr_sm.PredefinedDataType.drepr_uri.value
            else:
                assert node.datatype == O.LiteralNodeDataType.String
                datatype = drepr_sm.PredefinedDataType.xsd_string.value

            nodes[str(node.id)] = drepr_sm.LiteralNode(
                node_id=str(node.id), value=node.value, data_type=datatype
            )

    used_ids = {x for edge in sm.edges() for x in [str(edge.source), str(edge.target)]}
    for node_id in set(nodes.keys()).difference(used_ids):
        del nodes[node_id]

    for edge in sm.edges():
        edges[len(edges)] = drepr_sm.Edge(
            edge_id=len(edges),
            source_id=str(edge.source),
            target_id=str(edge.target),
            label=edge.rel_uri,
        )

    # print(edges)

    # add drepr:uri relationship
    for node in get_entity_data_nodes(sm, ident_props):
        new_node_id = str(node.id) + ":ents"
        nodes[new_node_id] = drepr_sm.DataNode(
            node_id=new_node_id,
            attr_id=get_ent_attr_id(node.col_index),
            data_type=drepr_sm.PredefinedDataType.drepr_uri.value,
        )
        inedges = [
            inedge for inedge in sm.in_edges(node.id) if inedge.abs_uri in ident_props
        ]
        assert len(inedges) == 1
        inedge = inedges[0]
        edges[len(edges)] = drepr_sm.Edge(
            edge_id=len(edges),
            source_id=str(inedge.source),
            target_id=new_node_id,
            # special predicate telling drepr to use as uri of entity, instead of generating a blank node
            label="drepr:uri",
        )

    prefixes = kgns_prefixes.copy()
    prefixes.update(drepr_sm.SemanticModel.get_default_prefixes())
    return drepr_sm.SemanticModel(
        nodes=nodes,
        edges=edges,
        prefixes=prefixes,
    )


def get_entity_data_nodes(
    sm: O.SemanticModel, ident_props: set[str]
) -> List[O.DataNode]:
    ent_dnodes = []
    for node in sm.iter_nodes():
        if not isinstance(node, O.DataNode):
            continue

        stypes = sm.get_semantic_types_of_column(node.col_index)
        if any(stype.predicate_abs_uri in ident_props for stype in stypes):
            ent_dnodes.append(node)

    return ent_dnodes



from __future__ import annotations

from collections import defaultdict
from io import BytesIO, StringIO
from typing import List, Mapping, Set

import orjson
import sm.outputs.semantic_model as O
from dependency_injector.wiring import Provide, inject
from drepr.models.prelude import (
    AlignedStep,
    Attr,
    CSVProp,
    DRepr,
    IndexExpr,
    Path,
    PMap,
    Preprocessing,
    PreprocessingType,
    RangeAlignment,
    RangeExpr,
    Resource,
    ResourceType,
)
from kgdata.misc.resource import RDFResource
from rdflib import RDF, Graph, URIRef
from sand_drepr.resources import get_entity_resource, get_table_resource
from sand_drepr.semanticmodel import get_drepr_sm, get_entity_data_nodes
from sand_drepr.transformation import get_transformation, has_transformation
from slugify import slugify
from sm.misc.funcs import assert_not_null
from sm.namespaces.prelude import KnowledgeGraphNamespace

from sand.config import AppConfig
from sand.extension_interface.export import IExport
from sand.helpers.namespace import NamespaceService
from sand.models.ontology import OntProperty, OntPropertyAR, OntPropertyDataType
from sand.models.table import Table, TableRow


def get_drepr_model(
    table_columns: list[str],
    table_size: int,
    sm: O.SemanticModel,
    kgns: NamespaceService,
    kgns_prefixes: dict[str, str],
    ontprop_ar: Mapping[str, OntProperty],
    ident_props: set[str],
) -> DRepr:
    """Create a D-REPR model of the dataset.

    Args:
        table_columns: list of column names
        table_size: number of rows in the table (exclude header)
        sm: the semantic model we want to convert
        ns: the graph namespace
        kgns_prefixes: the prefixes of the knowledge graph namespace
        ontprop_ar: mapping from the id to ontology property
        ident_props: list of properties that telling a data node contains entities (e.g., rdfs:label)
    """
    columns = [slugify(c).replace("-", "_") for c in table_columns]
    get_attr_id = lambda ci: f"{ci}__{columns[ci]}"
    get_ent_attr_id = lambda ci: f"{ci}__ent__{columns[ci]}"
    ent_dnodes = get_entity_data_nodes(sm, ident_props)

    attrs = [
        Attr(
            id=get_attr_id(ci),
            resource_id="table",
            path=Path(
                steps=[
                    RangeExpr(start=1, end=table_size + 1, step=1),
                    IndexExpr(val=ci),
                ]
            ),
            missing_values=[""],
        )
        for ci in range(len(table_columns))
    ]
    attrs += [
        Attr(
            id=get_ent_attr_id(node.col_index),
            resource_id="entity",
            path=Path(
                steps=[
                    RangeExpr(start=1, end=table_size + 1, step=1),
                    IndexExpr(val=node.col_index),
                ]
            ),
            missing_values=[""],
        )
        for node in ent_dnodes
    ]

    dsm = get_drepr_sm(
        sm=sm,
        kgns=kgns,
        kgns_prefixes=kgns_prefixes,
        ontprop_ar=ontprop_ar,
        ident_props=ident_props,
        get_attr_id=get_attr_id,
        get_ent_attr_id=get_ent_attr_id,
    )

    datatype_transformations = []
    for node in sm.nodes():
        if not isinstance(node, O.DataNode):
            continue

        datatypes: Set[OntPropertyDataType] = {
            assert_not_null(ontprop_ar[kgns.uri_to_id(inedge.abs_uri)]).datatype
            for inedge in sm.in_edges(node.id)
        }
        datatype = list(datatypes)[0] if len(datatypes) == 1 else None
        if datatype is None or not has_transformation(datatype):
            continue
        datatype_transformations.append(
            Preprocessing(
                type=PreprocessingType.pmap,
                value=PMap(
                    resource_id="table",
                    path=Path(
                        steps=[
                            RangeExpr(start=1, end=table_size + 1, step=1),
                            IndexExpr(val=node.col_index),
                        ]
                    ),
                    code=get_transformation(datatype),
                    change_structure=False,
                ),
            )
        )

    return DRepr(
        resources=[
            Resource(id="table", type=ResourceType.CSV, prop=CSVProp()),
            Resource(id="entity", type=ResourceType.CSV, prop=CSVProp()),
        ],
        preprocessing=datatype_transformations,
        attrs=attrs,
        aligns=[
            RangeAlignment(
                source=get_attr_id(0),
                target=get_attr_id(ci),
                aligned_steps=[AlignedStep(source_idx=0, target_idx=0)],
            )
            for ci in range(1, len(table_columns))
        ]
        + [
            RangeAlignment(
                source=get_attr_id(0),
                target=get_ent_attr_id(node.col_index),
                aligned_steps=[AlignedStep(source_idx=0, target_idx=0)],
            )
            for node in ent_dnodes
        ],
        sm=dsm,
    )



from pathlib import Path
from typing import Mapping

from sand.models.ontology import OntPropertyDataType

transdir = (Path(__file__).parent / "raw_transformations").absolute()

datatype2transformation: Mapping[OntPropertyDataType, Path] = {
    "globe-coordinate": (transdir / "global_coordinate.py"),
    "number": (transdir / "number.py"),
}

loaded_transformations: Mapping[OntPropertyDataType, str] = {}


def get_transformation(datatype: OntPropertyDataType):
    global loaded_transformations
    if datatype not in loaded_transformations:
        loaded_transformations[datatype] = datatype2transformation[datatype].read_text()
    return loaded_transformations[datatype]


def has_transformation(datatype: OntPropertyDataType):
    return datatype in datatype2transformation



from lat_lon_parser import parse

try:
    part = {" ", ".", "°", "′", "″", "'", '"'}
    norm_value = (
        value.lower().replace("b", "n").replace("đ", "e").replace("t", "w").upper()
    )
    split_index = -1
    for i, c in enumerate(norm_value):
        if c.isdigit() or c in part:
            continue
        split_index = i + 1
        break

    if split_index == -1:
        return value

    part1, part2 = norm_value[:split_index], norm_value[split_index:]
    if "E" in part2 or "W" in part2:
        lat, long = part1, part2
    else:
        lat, long = part2, part1

    lat = parse(lat)
    long = parse(long)
    return f"Point({long} {lat})"
except:
    return value



import csv
from io import StringIO
from typing import List, Set
from uuid import uuid4

from drepr.models import ResourceData, ResourceDataString

from sand.config import AppConfig
from sand.helpers.namespace import NamespaceService
from sand.models.table import Table, TableRow


def get_table_resource(table: Table, rows: List[TableRow]) -> ResourceData:
    """Return a CSV resource, in which the first row is the header"""
    f = StringIO()
    writer = csv.writer(
        f, delimiter=",", quoting=csv.QUOTE_MINIMAL, lineterminator="\n"
    )

    writer.writerow(table.columns)
    for row in rows:
        writer.writerow(row.row)

    return ResourceDataString(f.getvalue())


def get_entity_resource(
    appcfg: AppConfig,
    ns: NamespaceService,
    table: Table,
    rows: List[TableRow],
    ent_columns: Set[int],
) -> ResourceData:
    """Return a CSV resource of matrix mapping each position in a table to a corresponding entity uri."""
    new_entity_template: str = appcfg.entity.new_entity_template
    ent_rows: List[List[str]] = []
    kgns = ns.kgns

    for ri, row in enumerate(rows):
        ent_rows.append([])
        for ci in range(len(row.row)):
            if ci not in ent_columns:
                ent_rows[-1].append("")
                continue

            links = row.links.get(str(ci), [])
            # TODO: what happens when there are multiple entities?
            # for now, just return the first entity
            ent = None
            for link in links:
                if (
                    link.entity_id is not None
                    and link.entity_id != appcfg.entity.nil.id
                ):
                    ent = kgns.id_to_uri(link.entity_id)
                    break
            else:
                # generate new entity
                ent = new_entity_template.format(id=str(uuid4()))

            ent_rows[-1].append(ent)

    f = StringIO()
    writer = csv.writer(
        f, delimiter=",", quoting=csv.QUOTE_MINIMAL, lineterminator="\n"
    )

    writer.writerow(table.columns)
    for row in ent_rows:
        writer.writerow(row)
    return ResourceDataString(f.getvalue())



from __future__ import annotations

import re
from collections import defaultdict
from io import BytesIO, StringIO
from typing import List, Sequence, Set, cast

import orjson
import sm.outputs.semantic_model as O
from dependency_injector.wiring import Provide, inject
from drepr.main import convert
from drepr.models.prelude import (
    AlignedStep,
    Alignment,
    Attr,
    CSVProp,
    DRepr,
    IndexExpr,
    OutputFormat,
    Path,
    PMap,
    Preprocessing,
    PreprocessingType,
    RangeAlignment,
    RangeExpr,
    Resource,
    ResourceDataString,
    ResourceType,
)
from kgdata.misc.resource import RDFResource
from rdflib import RDF, Graph, URIRef
from sand.config import AppConfig
from sand.extension_interface.export import IExport
from sand.helpers.namespace import NamespaceService
from sand.models.ontology import OntPropertyAR, OntPropertyDataType
from sand.models.table import Table, TableRow
from slugify import slugify
from sm.misc.funcs import assert_isinstance, assert_not_null

from sand_drepr.resources import get_entity_resource, get_table_resource
from sand_drepr.semanticmodel import get_drepr_sm, get_entity_data_nodes
from sand_drepr.transformation import get_transformation, has_transformation


class DreprExport(IExport):
    @inject
    def __init__(
        self,
        appcfg: AppConfig = Provide["appcfg"],
        namespace: NamespaceService = Provide["namespace"],
        ontprop_ar: OntPropertyAR = Provide["properties"],
    ):
        self.appcfg = appcfg
        self.namespace = namespace
        self.ontprop_ar = ontprop_ar

    def export_data_model(self, table: Table, sm: O.SemanticModel) -> dict[str, str]:
        model = self.export_drepr_model(table, sm)
        return {
            "model.json": orjson.dumps(
                model.serialize(), option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS
            ).decode(),
            "model.yml": model.to_lang_yml(),
        }

    def export_extra_resources(
        self, table: Table, rows: list[TableRow], sm: O.SemanticModel
    ) -> dict[str, str]:
        if len(table.columns) == 0:
            # no column, no data
            return {}

        ent_columns = {
            node.col_index
            for node in get_entity_data_nodes(
                sm, self.appcfg.semantic_model.identifiers_set
            )
        }
        entresource = get_entity_resource(
            self.appcfg, self.namespace, table, rows, ent_columns
        )
        assert isinstance(entresource, ResourceDataString)
        return {
            "entity": (
                entresource.value.decode()
                if isinstance(entresource.value, bytes)
                else assert_isinstance(entresource.value, str)
            )
        }

    def export_data(
        self,
        table: Table,
        rows: List[TableRow],
        sm: O.SemanticModel,
        output_format: OutputFormat,
    ):
        """Convert a relational table into RDF format"""
        if len(table.columns) == 0:
            # no column, no data
            return ""

        ent_columns = {
            node.col_index
            for node in get_entity_data_nodes(
                sm, self.appcfg.semantic_model.identifiers_set
            )
        }
        resources = {
            "table": get_table_resource(table, rows),
            "entity": get_entity_resource(
                self.appcfg, self.namespace, table, rows, ent_columns
            ),
        }

        content = convert(
            repr=self.export_drepr_model(table, sm),
            resources=resources,
            format=output_format,
        )
        return self.post_processing(sm, content, output_format)

    def export_drepr_model(self, table: Table, sm: O.SemanticModel) -> DRepr:
        """Create a D-REPR model of the dataset."""
        columns = [slugify(c).replace("-", "_") for c in table.columns]

        existing_attr_names = {}

        def get_attr_id(ci):
            cname = slugify(columns[ci]).replace("-", "_")
            m = re.match(r"\d+([^\d].*)", cname)
            if m is not None:
                cname = m.group(1)
            if existing_attr_names.get(cname, None) != ci:
                return cname + "_" + str(ci)

            existing_attr_names[cname] = ci
            return cname

        get_ent_attr_id = lambda ci: f"{get_attr_id(ci)}__ent"
        ent_dnodes = get_entity_data_nodes(
            sm, self.appcfg.semantic_model.identifiers_set
        )

        attrs = [
            Attr(
                id=get_attr_id(ci),
                resource_id="table",
                path=Path(
                    steps=[
                        RangeExpr(start=1, end=table.size + 1, step=1),
                        IndexExpr(val=ci),
                    ]
                ),
                missing_values=[""],
            )
            for ci in range(len(table.columns))
        ]
        attrs += [
            Attr(
                id=get_ent_attr_id(node.col_index),
                resource_id="entity",
                path=Path(
                    steps=[
                        RangeExpr(start=1, end=table.size + 1, step=1),
                        IndexExpr(val=node.col_index),
                    ]
                ),
                missing_values=[""],
            )
            for node in ent_dnodes
        ]

        dsm = get_drepr_sm(
            sm=sm,
            kgns=self.namespace,
            kgns_prefixes=self.namespace.kgns_prefixes,
            ontprop_ar=self.ontprop_ar,
            ident_props=self.appcfg.semantic_model.identifiers_set,
            get_attr_id=get_attr_id,
            get_ent_attr_id=get_ent_attr_id,
        )

        datatype_transformations = []
        for node in sm.nodes():
            if not isinstance(node, O.DataNode):
                continue

            datatypes: Set[OntPropertyDataType] = {
                assert_not_null(self.ontprop_ar.get_by_uri(inedge.abs_uri)).datatype
                for inedge in sm.in_edges(node.id)
            }
            datatype = list(datatypes)[0] if len(datatypes) == 1 else None
            if datatype is None or not has_transformation(datatype):
                continue
            datatype_transformations.append(
                Preprocessing(
                    type=PreprocessingType.pmap,
                    value=PMap(
                        resource_id="table",
                        path=Path(
                            steps=[
                                RangeExpr(start=1, end=table.size + 1, step=1),
                                IndexExpr(val=node.col_index),
                            ]
                        ),
                        code=get_transformation(datatype),
                        change_structure=False,
                    ),
                )
            )

        aligns: list[Alignment] = []
        for ci in range(1, len(table.columns)):
            aligns.append(
                RangeAlignment(
                    source=get_attr_id(0),
                    target=get_attr_id(ci),
                    aligned_steps=[AlignedStep(source_idx=0, target_idx=0)],
                )
            )
        for node in ent_dnodes:
            aligns.append(
                RangeAlignment(
                    source=get_attr_id(0),
                    target=get_ent_attr_id(node.col_index),
                    aligned_steps=[AlignedStep(source_idx=0, target_idx=0)],
                )
            )

        return DRepr(
            resources=[
                Resource(id="table", type=ResourceType.CSV, prop=CSVProp()),
                Resource(id="entity", type=ResourceType.CSV, prop=CSVProp()),
            ],
            preprocessing=datatype_transformations,
            attrs=attrs,
            aligns=aligns,
            sm=dsm,
        )

    def post_processing(
        self, sm: O.SemanticModel, ttldata: str, output_format: OutputFormat
    ) -> str:
        """Post-processing the TTL data to fix until D-REPR addresses
        them.

        1. D-REPR doesn't generate relationships for literals that have outgoing edges to class nodes
        """
        outliterals = []
        for node in sm.iter_nodes():
            if isinstance(node, O.LiteralNode):
                if sm.out_degree(node.id) > 0:
                    outliterals.append(node)

        if len(outliterals) == 0:
            return ttldata

        assert output_format == OutputFormat.TTL, "Only support TTL output format"
        g = Graph()
        file = StringIO(ttldata)
        g.parse(file)

        source2triples = defaultdict(list)
        for s, p, o in g:
            source2triples[s].append((s, p, o))
        resources = [
            RDFResource.from_triples(k, ts) for k, ts in source2triples.items()
        ]

        new_triples = []
        for node in outliterals:
            node_value = URIRef(node.value)
            for edge in sm.out_edges(node.id):
                target_node = sm.get_node(edge.target)
                assert isinstance(target_node, O.ClassNode)
                for resource in resources:
                    if str(resource.props[str(RDF.type)][0]) == target_node.abs_uri:
                        new_triples.append(
                            (node_value, URIRef(edge.abs_uri), URIRef(resource.id))
                        )

        for triple in new_triples:
            g.add(triple)

        file = BytesIO()
        g.serialize(file, format="turtle")
        return file.getvalue().decode()
